import m114 algebra.pi_instances
set_option pp.beta true

def paire (f : ℝ → ℝ) := ∀ x, f (-x) = f x

def impaire (f : ℝ → ℝ) := ∀ x, f (-x) = -f x

-- Si f et g sont paires alors leur somme l'est aussi.
example (f g : ℝ → ℝ) : paire f → paire g →  paire (f + g) :=
begin
  -- Supposons que f est paire,
  intro hf,
  -- c'est à dire que, pour tout x, f(-x) = f(x)
  unfold paire at hf,
  -- et que g est paire.
  intro hg,
  -- c'est à dire que, pour tout x, g(-x) = g(x)
  unfold paire at hg,
  -- On doit montrer que, pour tout x, (f+g)(-x) = (f+g)(x)
  unfold paire,
  -- Soit x un réel.
  intro x,
  -- On calcule.
  calc (f + g) (-x) = f (-x) + g (-x) : by compute
  ... = f x + g (-x) : by rw hf x -- par hypothèse sur f (appliquée à x)
  ... = f x + g x : by rw hg x -- par hypothèse sur g (appliquée à x)
  ... = (f + g) x : by compute
end

/-
Dans la démonstration précédente, les lignes commençant par unfold
sont purement psychologiques, Lean n'en a pas besoin du tout.
De plus on peut faire plusieurs introduction d'un coup
en utilisant intros (au pluriel). On peut donc remplacer
les six premières lignes comme dans la version ci-dessous.

De plus Lean n'a pas vraiment besoin qu'on lui dise
à quel réel appliquer les hypothèses de parité, il
lui suffit de chercher dans le but en cours.
-/

-- Si f et g sont paires alors leur somme l'est aussi, avec une démonstration moins bavarde.
example (f g : ℝ → ℝ) : paire f → paire g →  paire (f + g) :=
begin
  -- Supposons que f et g sont paires et fixons un réel x.
  intros hf hg x,
  -- On calcule.
  calc (f + g) (-x) = f (-x) + g (-x) : by compute
  ... = f x + g (-x) : by rw hf -- par hypothèse sur f
  ... = f x + g x : by rw hg -- par hypothèse sur g
  ... = (f + g) x : by compute
end

example (f g : ℝ → ℝ) : paire f → paire (g ∘ f) :=
begin
  sorry
end

example (f g : ℝ → ℝ) : impaire f → impaire g →  impaire (g ∘ f) :=
begin
  sorry
end


def croissante (f : ℝ → ℝ) := ∀ x₁ x₂, x₁ ≤ x₂ → f x₁ ≤ f x₂

def decroissante (f : ℝ → ℝ) := ∀ x₁ x₂, x₁ ≤ x₂ → f x₁ ≥ f x₂

example (f g : ℝ → ℝ) (hf : croissante f) (hg : croissante g) : croissante (g ∘ f) :=
begin
  -- Soit x₁ et x₂ des réels tels que x₁ ≤ x₂
  intros x₁ x₂ h,
  -- Comme f est croissante, f x₁ ≤ f x₂.
  have etape1 :  f x₁ ≤ f x₂,
    exact hf x₁ x₂ h,
  -- Comme g est croissante, on en déduit g(f x₁) ≤ g(f x₂).
  exact hg (f x₁) (f x₂) etape1,
end

/-
On peut aussi utiliser la commande specialize pour spécialiser un énoncé quantifié
universellement, comme dans la variante suivante.
-/
example (f g : ℝ → ℝ) (hf : croissante f) (hg : croissante g) : croissante (g ∘ f) :=
begin
  -- Soit x₁ et x₂ des réels tels que x₁ ≤ x₂
  intros x₁ x₂ h,
  -- L'hypothèse de croissance sur f implique en particulier que f x₁ ≤ f x₂.
  specialize hf x₁ x₂ h,
  -- Comme g est croissante, on en déduit g(f x₁) ≤ g(f x₂).
  exact hg (f x₁) (f x₂) hf,
end

-- Pourriez-vous remplacer les deux dernières lignes de la démonstration précédente
-- par une seule ligne ?

-- Le même en raisonnant vers l'arrière. On remarquera que Lean se débrouille pour comprendre à
-- quels nombres réels appliquer les hypothèses de croissance.
example (f g : ℝ → ℝ) (hf : croissante f) (hg : croissante g) : croissante (g ∘ f) :=
begin
  -- Soit x₁ et x₂ des réels tels que x₁ ≤ x₂
  intros x₁ x₂ h,
  -- Il faut montrer (g ∘ f) x₁ ≤ (g ∘ f) x₂.
  -- Comme g est croissante, il suffit de montrer f x₁ ≤ f x₂
  apply hg,
  -- ce qui découle de l'hypothèse sur f
  apply hf,
  -- et de notre hypothèse sur x₁ et x₂
  exact h
end

example (f g : ℝ → ℝ) (hf : croissante f) (hg : decroissante g) : decroissante (g ∘ f) :=
begin
  sorry
end

/-
Le symbole ∨ désigne le connecteur logique « ou ».
Pour utiliser une hypothèse de la forme
hyp : P ∨ Q
on tape
cases hyp with hP hQ
qui créé deux branches dans la démonstration :
une branche où hyp est remplacée par hP : P
et l'autre dans laquelle elle est remplacée par hQ : Q.

Pour démontrer un but de la forme P ∨ Q,
on utilise left et on démontre P ou right
et on démontre Q.

Dans l'exemple suivant, on utilise aussi
la commande linarith qui cherche à démontrer
une égalité ou une inégalité par combinaison
linéaire des hypothèses. Cette commande
permet aussi de déduire a ≤ b de a < b
ou d'autres trivialités de ce genre.

Enfin on utilise le lemme
mul_eq_zero : a*b = 0 ↔ a = 0 ∨ b = 0
qui servira aussi dans l'exercice suivant.
-/

example (a b : ℝ) : a = a*b → a = 0 ∨ b = 1 :=
begin
  intro hyp,
  have H : a*(1 - b) = 0,
    calc a*(1 - b) = a - a*b : by compute
               ... = 0       : by linarith,
  rw mul_eq_zero at H,
  cases H with Ha Hb,
    left,
    exact Ha,
  right,
  linarith,
end

example (x y : ℝ) : x^2 = y^2 → x = y ∨ x = -y :=
begin
  sorry
end

/-
Dans l'exercice suivant, on pourra utiliser le lemme
eq_or_lt_of_le : x ≤ y → x = y ∨ x < y
-/
example (f : ℝ → ℝ) : croissante f ↔ ∀ x y, x < y → f x ≤ f y :=
begin
  sorry
end

/-
Dans l'exercice suivant, on pourra utiliser le lemme
le_total x y : x ≤ y ∨ y ≤ x
-/
example (f : ℝ → ℝ) (h : croissante f) (h' : ∀ x, f (f x) = x) : ∀ x, f x = x :=
begin
  sorry
end

